DEFAULT_AUTH_PATH = "/api/rest/v1/auth"

DEFAULT_JSON_API_PATH = "/api/json"
DEFAULT_JSON_API_VERSION = "v1"

DEFAULT_DEV_OPS_URL = "api.astra.datastax.com"
DEFAULT_DEV_OPS_API_VERSION = "v2"
DEFAULT_DEV_OPS_AUTH_HEADER = "Authorization"

DEFAULT_TIMEOUT = 30000
DEFAULT_AUTH_HEADER = "Token"
DEFAULT_KEYSPACE_NAME = "default_keyspace"
DEFAULT_REGION = "us-east1"

MAX_INSERT_NUM_DOCUMENTS = 20
